unit UnitFrmGestures;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Menus, Buttons,  UnitFrmMouseMonitor,
  XPMan, StdCtrls {for THEME support}, UnitFrmDummyUnicodeTooltip, ExtCtrls;
type TActionType = (atKeystroke, atMultiKeystroke, atProgram);
type
  TFrmGestures = class(TForm)
    lvGestures: TListView;
    SpeedButton1: TSpeedButton;
    pmRightClick: TPopupMenu;
    Edit1: TMenuItem;
    Delete1: TMenuItem;
    EditGesture1: TMenuItem;
    pmNew: TPopupMenu;
    NewKeystroke1: TMenuItem;
    MultiKeystrokesTrigger1: TMenuItem;
    pmDirections: TPopupMenu;
    GestMenu: TMenuItem;
    StaticText1: TStaticText;
    SpeedButton3: TSpeedButton;
    imgPreview: TImage;
    cbxSelect: TComboBoxEx;
    timCloseTooltip: TTimer;
    timHelpConfig: TTimer;
    Runfileprogram1: TMenuItem;
    N1: TMenuItem;
    cbDisableAll: TCheckBox;


    procedure Edit1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure SpeedButton1Click(Sender: TObject);
    procedure NewKeystroke1Click(Sender: TObject);
    procedure EditGesture1Click(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure Delete1Click(Sender: TObject);
    procedure MultiKeystrokesTrigger1Click(Sender: TObject);

    procedure SpeedButton2Click(Sender: TObject);
    procedure SpeedButton3Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure cbxSelectDropDown(Sender: TObject);
    procedure cbxSelectClick(Sender: TObject);
    procedure cbxSelectSelect(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure timCloseTooltipTimer(Sender: TObject);
    procedure lvGesturesClick(Sender: TObject);

    procedure GestMenuClick(Sender: TObject);
    procedure timHelpConfigTimer(Sender: TObject);
    procedure Runfileprogram1Click(Sender: TObject);
  private
    { Private declarations }
    Keystrokes : TStringList;
    ActionTypes : TStringList;
    Gestures : TStringList;
    GlobalKeystrokes : TStringList;
    GlobalGestures : TStringList;
    GlobalActionTypes : TStringList;
    EnableSuggestions : boolean;
    AutoCorrection : boolean;

    ExecutePause : integer;

    DoExit : boolean;
    tt : TTooltipWindow;
    lastINI : String;
    function toString(at : TActionType) : string; overload;
    function fromString(s : string) : TActionType; overload;
    procedure SaveGestures(filename : string = 'Gestures.ini');
    procedure LoadGestures(filename : string = 'Gestures.ini'; SkipGUI : boolean = false);


    procedure HelpConfigureClick(Sender: TObject);
    procedure SuggestionConfigureClick(Sender: TObject);
    function EXEtoINI(exe : string) : string;
    procedure RunGesture(gesture : string);
    procedure LoadGestureFromWizard(NewItem : boolean = true);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    function ToHumanReadable(s : string) : string;
    procedure ReportGestures(gesture : string; NoTooltip : boolean = false);
    procedure ReportGesturesDiagonal(gesture : string);
    procedure ReportFirstDirection(md : TMouseDirection);
    { Public declarations }
    procedure ExitProgram;
    procedure WMEndSession(var msg : TWMQueryEndSession); message WM_QUERYENDSESSION;
    function ProgramIsDisabled : boolean;

    property prEnableSuggestions : boolean read EnableSuggestions write EnableSuggestions;
    property prEnabledAutoCorrection : boolean read AutoCorrection write AutoCorrection;
    property prExecutePause : integer read ExecutePause  write ExecutePause;
  end;

var
  FrmGestures: TFrmGestures;


implementation

uses UnitMisc, UnitFrmHotKey, IniFiles, UnitFrmGestureEdit,
  UnitKeyboardQuery, UnitPaste,
  UnitFrmKeystrokeList, UnitToken, UnitFrmConfig, UnitDrawGesture,
  UnitFrmSysTrayMenu, UnitFrmScratchpad, UnitFrmProgramEdit, StrUtils,
  UnitFrmGestureWizard, UnitFrameHotKey;

{$R *.dfm}


procedure TFrmGestures.FormClick(Sender: TObject);
begin
    if KeyboardQuery.IsPressed(VK_SHIFT) then begin
        FrmMouseMonitor.show;
    end;

end;

procedure TFrmGestures.FormClose(Sender: TObject; var Action: TCloseAction);
begin
    //self.SaveGestures(lastINI);
end;

procedure TFrmGestures.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
    self.SaveGestures(lastINI);
    if not DoExit then  begin
        self.hide;
    end;
    CanClose := DoExit;
end;

procedure TFrmGestures.FormCreate(Sender: TObject);
var kd : TKeystrokeData;
    j : integer;
begin
    Keystrokes := TStringList.Create;
    Gestures := TStringList.Create;
    ActionTypes := TStringList.Create;
    GlobalKeystrokes := TStringList.Create;
    GlobalGestures := TStringList.Create;
    GlobalActionTypes := TStringList.Create;
    tt := TTooltipWindow.Create();
    LoadGestures();


    cbxSelectDropDown(self);
    cbxSelect.ItemIndex := 0;
end;




procedure TFrmGestures.FormShow(Sender: TObject);
begin
    self.cbxSelectSelect(sender); 
end;

procedure TFrmGestures.cbxSelectClick(Sender: TObject);
var sr : TSearchRec;
    i : integer;
begin
end;

procedure TFrmGestures.cbxSelectDropDown(Sender: TObject);
var sr : TSearchRec;
    i, idx : integer;
    h : THandle;
    s, ini : string;
begin
    if cbxSelect.items.indexof('Gestures.ini') = -1 then begin
        cbxSelect.Items.Add('Gestures.ini');
        cbxSelect.ItemsEx[cbxSelect.ItemsEx.Count-1].ImageIndex := 6;
    end;
    // add only new found INI files
    i := FindFirst(UnitMisc.GetAppPath + '\Gesture*.ini',faAnyFile, sr);
    while i=0 do begin
        if (sr.Attr and faDirectory) = 0 then begin
            idx := cbxSelect.Items.IndexOf(sr.name);
            if idx = -1 then begin
                cbxSelect.Items.Add(sr.name);
                cbxSelect.ItemsEx[cbxSelect.ItemsEx.Count-1].ImageIndex := 6;
            end else begin
                cbxSelect.ItemsEx[idx].ImageIndex := 6;
            end;
        end;

        i := FindNext(sr);
    end;
    FindClose(sr);

    h := Windows.GetForegroundWindow;
    while h <> 0 do begin
        s := WindowHandleToEXEName(h);
        if lowercase(s) <> 'mousewrangler.exe' then begin
            ini := EXEtoINI(s);
            idx := cbxSelect.Items.IndexOf(ini);
            if idx = -1 then begin
                cbxSelect.Items.Add(ini);
                cbxSelect.ItemsEx[cbxSelect.ItemsEx.Count-1].ImageIndex := -1;
            end else begin
                //cbxSelect.ItemsEx[idx].ImageIndex := -1;
            end;
        end;
        h := GetNextWindow(h);
    end;



end;

procedure TFrmGestures.cbxSelectSelect(Sender: TObject);
begin
    cbDisableAll.Visible := false;
    if cbxSelect.Items[cbxselect.itemindex] <> 'Gestures.ini' then begin
        cbDisableAll.visible := true;
    end;
    cbDisableAll.caption := 'Disable all gestures for ' +
        replacestr(replacestr(cbxselect.items[cbxselect.itemindex],'Gestures_',''),
            '.ini','')
    ;
    
    if cbxSelect.ItemIndex = 0 then begin
        cbxSelect.Hint := 'Global Gestures  [Gestures.ini]';
    end else begin
        cbxSelect.Hint := 'Program Specific Gestures  ['+ cbxSelect.Items[cbxSelect.itemIndex] +']';
    end;
    self.SaveGestures(lastINI);
    Self.LoadGestures(cbxSelect.Items[cbxSelect.itemIndex]);
end;

procedure TFrmGestures.CreateParams(var Params: TCreateParams);
begin
    inherited CreateParams(Params);
    with Params do begin
        ExStyle := ExStyle or WS_EX_APPWINDOW and not WS_EX_TOOLWINDOW ;
        WndParent := GetDesktopwindow;
    end;
end;

procedure TFrmGestures.Delete1Click(Sender: TObject);
begin
    if lvGestures.selected = nil then EXIT;


    Gestures.values[lvGestures.Selected.caption] := '';
    Keystrokes.values[lvGestures.Selected.caption] := '';
    ActionTypes.values[lvGestures.Selected.caption] := '';
    lvGestures.Items.Delete(lvGestures.Selected.Index);

    self.SaveGestures(self.lastINI);
end;

procedure TFrmGestures.Edit1Click(Sender: TObject);
var kd : TKeystrokeData;
    sl : TStringList;
    i : integer;
    s, data, prog, param : string;
begin
    if lvGestures.selected = nil then EXIT;

    FrmGestureWizard.Gesture := Gestures.Values[lvGestures.Selected.Caption];
    FrmGestureWizard.ActionType := FromString(ActionTypes.values[FrmGestureWizard.Gesture]);
    FrmGestureWizard.ActionData := Keystrokes.values[FrmGestureWizard.Gesture];

    if FrmGestureWizard.EditGesture(true) then begin
        self.LoadGestureFromWizard(False); 
    end;

     {
    case fromstring(lvGestures.Selected.SubItems[0]) of
    atKeystroke: begin

        kd := FrmKeystrokeEdit.FromString( self.Keystrokes.values[lvGestures.Selected.SubItems[1]] );
        if FrmKeystrokeEdit.GetHotkey(kd, true) then begin
            Keystrokes.Values[Gestures.Values[lvGestures.Selected.caption]] := FrmKeystrokeEdit.ToString(kd);
            lvGestures.Selected.SubItems[1] := kd.name;
        end;
    end;
    atMultiKeystroke: begin
        sl := TStringList.Create;
        if FrmKeystrokeList.GetKeystrokeList(sl, Keystrokes.Values[ Gestures.Values[lvGestures.Selected.caption] ] ) then begin
            s := '';
            for i := 0 to sl.count - 1 do begin
                kd := FrmKeystrokeEdit.FromString(sl.Strings[i]);
                if (s <> '') then s := s + ' | ';
                s := s + kd.name;
                if data <> '' then data := data + '|';
                data := data +  sl.Strings[i];

                //Keystrokes.Values[Gestures.Values[lvGestures.Selected.caption]] := data;
            end;
            Keystrokes.Values[ Gestures.Values[lvGestures.Selected.caption]] := data;
            lvGestures.selected.SubItems[1] := s;
        end;
    end;
    atProgram: begin
        if FrmProgramEdit.GetProgram(prog, param) then begin
            Keystrokes.values[ Gestures.Values[lvGestures.Selected.caption]] := prog +'|'+param;
            lvGestures.selected.SubItems[1] := prog + ' | ' +param;;
        end;
    end;
    end;}
end;

procedure TFrmGestures.EditGesture1Click(Sender: TObject);
var name, oldgesture, gesture, key, actiontype: string;
begin
    if lvGestures.selected = nil then EXIT;

    FrmGestureWizard.Gesture := Gestures.Values[lvGestures.Selected.Caption];
    FrmGestureWizard.ActionType := FromString(ActionTypes.values[FrmGestureWizard.Gesture]);
    FrmGestureWizard.ActionData := Keystrokes.values[FrmGestureWizard.Gesture];

    if FrmGestureWizard.EditGesture then begin
        self.LoadGestureFromWizard(False); 
    end;
    {
    if FrmGestureEdit.GetGesture(name, gesture) then begin
        try
            // save the old gesture
            oldgesture := Gestures.Values[lvGestures.selected.caption];
            Gestures.Values[lvGestures.selected.caption] := '';
            key := Keystrokes.Values[oldgesture];
            Keystrokes.Values[oldgesture] := '';
            actiontype := ActionTypes.Values[oldgesture];
            ActionTypes.Values[oldgesture] := '';

            //  restore setting using the new gesture
            lvGestures.Selected.caption := name;
            Gestures.Values[name] := gesture;
            Keystrokes.Values[gesture] := key;
            ActionTypes.Values[gesture] := actiontype;
        except
            // ignore if it's no longer selected
        end;
    end;}
end;

function TFrmGestures.EXEtoINI(exe: string): string;
begin
    result := 'Gestures_' + exe + '.ini';
end;

procedure TFrmGestures.ExitProgram;
begin
    self.DoExit := true;
    self.close;
    Application.Terminate;
end;

procedure TFrmGestures.WMEndSession(var msg : TWMQueryEndSession);
begin

    msg.Result := 1;       // do allow close
    self.ExitProgram;
end;



procedure TFrmGestures.ReportFirstDirection(md: TMouseDirection);
var i : integer;
    li : TListItem;
    mi, submi : TMenuItem;
    s, find : string;
    h : thandle;

    sl : TStringList;

    procedure AddMenu(gest : string; Gestures, ActionTypes, Keystrokes : TStringList; index : integer = -1);
    var kd : TKeystrokeData;
        s : string;
        keys, keyname : string;
    begin
        s :=  Gestures.Values[ ToHumanReadable(gest)];
        if s <>'' then begin
            //if TMouseDirection(StrToInt(s[1])) <> TMouseDirection(StrToInt(gesture[1])) then CONTINUE;


            keys := Keystrokes.Values[gest];
            while keys <> '' do begin
                s := TokenString(keys,'|');

                case FromString(ActionTypes.Values[gest]) of
                    atKeystroke, atMultiKeystroke: begin
                        kd := FrmKeystrokeEdit.FromString(s);
                        if keyname <> '' then keyname := keyname + ' | ';
                        keyname := keyname + kd.name;
                    end;
                    atProgram: begin
                        keyname := ExtractFileName(s);
                        tokenstring(keys,'|');
                    end;
                end;



            end;

            mi := pmDirections.CreateMenuItem;
            case FromString(ActionTypes.Values[gest]) of
                atKeystroke, atMultiKeystroke: mi.caption := ToHumanReadable(gest) + #9+  '   ('+ keyname   + ')' ;
                atProgram: mi.caption := ToHumanReadable(gest) +  #9 + '   ['+ keyname   + ']' ;
            end;



            mi.hint := gest;
            mi.OnClick := self.GestMenuClick;
            pmDirections.Items.Add(mi);
        end;
    end;
begin
    pmDirections.items.clear;



    s := WindowHandleToEXEName(GetForegroundWindow);
    s := EXEtoINI(s);
    self.LoadGestures(s,TRUE);


    for i := 0 to Gestures.count - 1 do begin
        s :=  Gestures.ValueFromIndex[i];
        if s = '' then CONTINUE;
        if TMouseDirection(StrToInt(s[1])) <> md then CONTINUE;
        AddMenu(s, Gestures, ActionTypes, Keystrokes);
        //AddMenu(s, GlobalGestures, GlobalKeystrokes);
    end;
    for i := 0 to GlobalGestures.count - 1 do begin
        s :=  GlobalGestures.ValueFromIndex[i];
        if s = '' then CONTINUE;
        if TMouseDirection(StrToInt(s[1])) <> md then CONTINUE;
        //AddMenu(s, Gestures, Keystrokes);
        AddMenu(s, GlobalGestures, GlobalActionTypes, GlobalKeystrokes);
    end;



    if pmDirections.Items.count > 0 then begin

        Tt.ShowTooltip('Help Popup  ', point(mouse.CursorPos.X + 15, mouse.cursorpos.y),false,'');
        KeyboardQuery.WaitUntilRelease(RightButton);
        tt.CloseTooltip;

        FrmMouseMonitor.AbortInterpret;
        FrmMouseMonitor.SetMonitorEnabled(false);

        mi := pmDirections.CreateMenuItem;
        mi.Caption := '-';
        pmDirections.Items.add(mi);
        mi := pmDirections.CreateMenuItem;
        mi.Caption := 'Close';
        pmDirections.Items.add(mi);

        mi := pmDirections.CreateMenuItem;
        mi.Caption := 'Help Popup';
        submi := pmDirections.CreateMenuItem;
        submi.Caption := 'Configure';
        submi.OnClick := self.HelpConfigureClick;
        mi.Add(submi);
        pmDirections.Items.insert(0,mi);
        mi := pmDirections.CreateMenuItem;
        mi.Caption := '-';
        pmDirections.Items.insert(1,mi);


        FrmMouseMonitor.ClosePopups;
        h := FrmScratchPad.RemoveSelfFromFocus(True);
        if threadattach(GetForegroundWindow) then begin
            pmDirections.Popup(mouse.CursorPos.X - 60, mouse.CursorPos.Y + 20);
            sleep(10);

            SetForegroundWindow(h);

            pmDirections.Popup(mouse.CursorPos.X - 60, mouse.CursorPos.Y + 20);

            ThreadDetach;
        end;
        FrmMouseMonitor.SetMonitorEnabled(true);

    end else begin
        Tt.ShowTooltip('Help Popup - no matches', point(mouse.CursorPos.X + 15, mouse.cursorpos.y),false,'');
        KeyboardQuery.WaitUntilRelease(RightButton);
        tt.CloseTooltip;
    end;
end;

function TFrmGestures.ProgramIsDisabled : boolean;
var s : string;
    x : TIniFile;
begin
    s := WindowHandleToEXEName(GetForegroundWindow);
    s := EXEtoINI(s);
    x := TInifile.create(UnitMisc.GetAppPath + s);
    result := x.ReadBool('settings',cbDisableAll.Name, false);

    x.Free;
end;

procedure TFrmGestures.ReportGestures(gesture: string; NoTooltip : boolean = false);
var name, s, data, actiontype : string;
    li : TListItem;
    kd : TKeystrokeData;
    //tt : TTooltipWindow;
    p : Tpoint;
    i : integer;
    found : boolean;
    b : boolean;
begin
    name := ToHumanReadable(gesture);

    // load the per-process specific INI
    s := WindowHandleToEXEName(GetForegroundWindow);
    s := EXEtoINI(s);
    self.LoadGestures(s,TRUE);

    if cbdisableall.checked and (not NoTooltip) then begin
        p := Mouse.CursorPos;
        p.X := p.x + 11;
        tt.ShowTooltip('Disabled ', p,false,'');
        timCloseTooltip.Enabled := true;;
        EXIT;
    end;


    b := FrmScratchPad.Visible;

    //FrmScratchPad.Visible := false;
    //Application.ProcessMessages;

    found := false;
    if Gestures.Values[name] <> '' then begin
        found := true;
        self.RunGesture(Gestures.Values[name]);
    end else if GlobalGestures.values[name] <> '' then begin
        found := true;
        self.RunGesture(GlobalGestures.Values[name]);
    end;

    if found and (not NoTooltip) then begin
        //tt := TTooltipWindow.create;
        p := Mouse.CursorPos;
        p.X := p.x + 11;
        tt.ShowTooltip(name + '  ', p,false,'');
        timCloseTooltip.Enabled := true;;
    end;

    sleep(200);
    FrmScratchPad.Visible := b;
end;

procedure TFrmGestures.ReportGesturesDiagonal(gesture: string);
var name, s, data, actiontype : string;
    li : TListItem;
    p : Tpoint;
    i, errx : integer;
    found : boolean;
    sim, simend : TStringList;

    procedure AddSim(sim : TStringList; gest : string; index : integer = -1);
    begin
        if sim.IndexOf(gest) = -1 then begin
            if index = -1 then begin
                sim.Add(gest);
            end else begin
                sim.Insert(index, gest);
            end;
        end;
    end;

    function GetSimilarGestures(Gestures,ActionTypes,Keystrokes : TStringList) : boolean;
    var i,j, x : integer;
        gest : string;
    begin
        for i := 0 to Gestures.count - 1 do begin
            gest := Gestures.ValueFromIndex[i];
            x := 0;

            if length(gest)=length(gesture) then begin
                for j := 1 to length(gesture) do begin
                    if gest[j] = gesture[j] then inc(x);
                end;

                if (length(gesture)=2) and (x >= trunc(length(gesture)  * 0.5)) or
                    (length(gesture)=3) and (x >= trunc(length(gesture)  * 0.67)) or
                    (length(gesture)>=4) and (x >= trunc(length(gesture)  * 0.60)) then begin
                    if gesture[1] = gest[1] then begin
                        AddSim(sim, gest, 0);
                    end else begin
                        AddSim(simend, gest);
                    end;
                end;
            end;
        end;
    end;

    function GetSimilarGesturesByDelete(Gestures,ActionTypes,Keystrokes : TStringList; index : integer = 0) : boolean;
    var i,j, x : integer;
        gest, oneoff : string;
    begin
        // gesture has +1 movement
        for i := 0 to Gestures.count - 1 do begin
            gest := Gestures.ValueFromIndex[i];
            x := 0;


            for j := 1 to length(gesture) do begin
                oneoff := gesture;
                delete(oneoff,j,1);

                if oneoff = gest then
                    if gesture[1] = gest[1] then begin
                        AddSim(sim, gest, index);
                    end else begin
                        AddSim(simend, gest);
                    end;
            end;
        end;

        // gesture has -1 movement
        if length(gesture) > 1 then
            for i := 0 to Gestures.count - 1 do begin
                gest := Gestures.ValueFromIndex[i];
                x := pos(gesture, gest);

                if (x <> 0) and (length(gest) = length(gesture)+1) then begin
                    if x = 1 then begin
                        AddSim(sim, gest, index);
                    end else begin
                        AddSim(simend, gest);
                    end;
                end;
            end;

    end;

    function FoundGestures(name : string; Gestures,ActionTypes,Keystrokes : TStringList) : boolean;
    var kd : TKeystrokeData;
        h : THandle;
    begin
        result := false;
        if Gestures.Values[name] <> '' then begin
            self.RunGesture(Gestures.Values[name]);
            result := true;
        end;
    end;

    function FoundCorrection(sim : TStringList) : boolean;
    var k, j : integer;
    begin
        result := false;
        for k := 0 to sim.count - 1 do begin
            for j := 1 to length(gesture) do begin
                if sim.Strings[k][j] <> gesture[j] then begin
                    with FrmMouseMonitor do begin
                        if IsBetweenNeighbors(
                            TMouseDirection(StrToInt(sim.Strings[k][j])), TMouseDirection(StrToInt(gesture[j])),
                            GetDirectionAngle(j-1)
                            )  then begin

                            self.RunGesture(sim.Strings[k]);
                            result := true;

                            EXIT;
                        end;
                    end;

                end;
            end;
        end;
    end;
var mi, submi : TMenuItem;

    procedure AddMenu(gest : string; Gestures, ActionTypes, Keystrokes : TStringList; index : integer = -1);
    var kd : TKeystrokeData;
        s : string;
        keys, keyname : string;
    begin
        s :=  Gestures.Values[ ToHumanReadable(gest)];
        if s <>'' then begin
            //if TMouseDirection(StrToInt(s[1])) <> TMouseDirection(StrToInt(gesture[1])) then CONTINUE;


            keys := Keystrokes.Values[gest];
            while keys <> '' do begin
                s := TokenString(keys,'|');

                case FromString(ActionTypes.Values[gest]) of
                    atKeystroke, atMultiKeystroke: begin
                        kd := FrmKeystrokeEdit.FromString(s);
                        if keyname <> '' then keyname := keyname + ' | ';
                        keyname := keyname + kd.name;
                    end;
                    atProgram: begin
                        keyname := ExtractFileName(s);
                        tokenstring(keys,'|');
                    end;
                end;
            end;

            mi := pmDirections.CreateMenuItem;
            case FromString(ActionTypes.Values[gest]) of
                atKeystroke, atMultiKeystroke: mi.caption := ToHumanReadable(gest) + #9+  '   ('+ keyname   + ')' ;
                atProgram: mi.caption := ToHumanReadable(gest) +  #9 + '   ['+ keyname   + ']' ;
            end;



            mi.hint := gest;
            mi.OnClick := self.GestMenuClick;
            pmDirections.Items.Add(mi);
        end;
    end;
var h : THandle;
    j, k : integer;
    kd : TKeystrokeData;
    idx : integer;
    b : boolean;


begin
    name := ToHumanReadable(gesture);

    // load the per-process specific INI
    s := WindowHandleToEXEName(GetForegroundWindow);
    s := EXEtoINI(s);
    self.LoadGestures(s,TRUE);


    b := FrmScratchPad.Visible;
    //FrmScratchPad.Visible := false;
    FrmScratchPad.RemoveSelfFromFocus;

    if (not FoundGestures(name, Gestures,ActionTypes,Keystrokes)) and
        (not FoundGestures(name, GlobalGestures,GlobalActionTypes,GlobalKeystrokes)) then begin

        if self.EnableSuggestions  then begin
            sim := TStringList.Create;
            simend := TStringList.Create;
            GetSimilarGestures(Gestures,ActionTypes,Keystrokes);
            GetSimilarGestures(GlobalGestures,GlobalActionTypes,GlobalKeystrokes);

            if self.AutoCorrection  then begin
                if FoundCorrection(sim) or FoundCorrection(simend)  then begin
                    sleep(50);
                    p := Mouse.CursorPos;
                    p.X := p.x + 11;
                    tt.ShowTooltip(name + '  ', p,false,'');
                    timCloseTooltip.Enabled := true;;
                    EXIT;
                end;
            end;

            pmDirections.items.clear;

            if sim.count > 2 then begin
                sim.AddStrings(simend);
                for i := 0 to Sim.count - 1 do begin
                    AddMenu(sim.strings[i], Gestures, ActionTypes, Keystrokes);
                    AddMenu(sim.strings[i], GlobalGestures,GlobalActionTypes, GlobalKeystrokes);
                end;
            end else begin
                idx := 0;
                GetSimilarGesturesByDelete(Gestures,ActionTypes,Keystrokes, idx);
                GetSimilarGesturesByDelete(GlobalGestures,GlobalActionTypes,GlobalKeystrokes, idx);

                sim.AddStrings(simend);
                for i := 0 to Sim.count - 1 do begin
                    AddMenu(sim.strings[i], Gestures,ActionTypes, Keystrokes, idx);
                    AddMenu(sim.strings[i], GlobalGestures,GlobalActionTypes, GlobalKeystrokes, idx);
                end;
            end;

            if pmDirections.Items.count > 0 then begin
                FrmMouseMonitor.AbortInterpret;
                FrmMouseMonitor.SetMonitorEnabled(false);

                mi := pmDirections.CreateMenuItem;
                mi.Caption := '-';
                pmDirections.Items.add(mi);
                mi := pmDirections.CreateMenuItem;
                mi.Caption := 'Close';
                pmDirections.Items.add(mi);

                mi := pmDirections.CreateMenuItem;
                mi.Caption := 'Suggestions Popup ' ;
                submi := pmDirections.CreateMenuItem;
                submi.Caption := 'Configure ';
                submi.OnClick := self.SuggestionConfigureClick;
                mi.Add(submi);
                pmDirections.Items.Insert(0,mi);

                mi := pmDirections.CreateMenuItem;
                mi.Caption := '('+ToHumanReadable(gesture) + ')  not found ' ;
                mi.Enabled := false;
                pmDirections.Items.Insert(1,mi);

                mi := pmDirections.CreateMenuItem;
                mi.Caption := '-';
                pmDirections.Items.Insert(2,mi);

                FrmMouseMonitor.ClosePopups;
                h := FrmScratchPad.RemoveSelfFromFocus(True);
                if threadattach(GetForegroundWindow) then begin
                    pmDirections.Popup(mouse.CursorPos.X - 60, mouse.CursorPos.Y + 20);
                    sleep(10);
                    pmDirections.Popup(mouse.CursorPos.X - 60, mouse.CursorPos.Y + 20);
                    SetForegroundWindow(h);
                    ThreadDetach;
                end;
                FrmMouseMonitor.SetMonitorEnabled(true);
            end;

            myfree(sim);
            myfree(simend);
        end;
    end else begin
        //tt := TTooltipWindow.create;
        sleep(50);
        p := Mouse.CursorPos;
        p.X := p.x + 11;
        tt.ShowTooltip(name + '  ', p,false,'');
        timCloseTooltip.Enabled := true;;
    end;
    sleep(50);
    FrmScratchPad.Visible := b;
end;





procedure TFrmGestures.LoadGestures(filename : string = 'Gestures.ini'; SkipGUI : Boolean = false);
var kd : TKeystrokeData;
  name: string;
  ini: TIniFile;
  sl: TStringList;
  i: Integer;
  li: TListItem;
  Gest, GestName, ActionType, Action ,s : string;
  display : string;
  IsInt : boolean;
begin
    if not SKipGUI then begin
        lvGestures.Clear;
    end;
    Gestures.Clear;
    Keystrokes.Clear;
    lastINI := filename;

    name := UnitMisc.GetAppPath + filename;
    cbDisableAll.Checked := false;
    if FileExists(name) then begin
        ini := TIniFile.create(name);

        cbDisableAll.Checked := ini.ReadBool('Settings', cbDisableAll.Name, false);

        sl := TStringList.Create;
        ini.ReadSections(sl);
        for i := 0 to sl.count - 1 do begin
            Gest := sl.Strings[i];
            try
                StrToInt(gest);
                IsInt := true;
            except
                IsInt := false;
            end;
            if not IsInt then CONTINUE;
            

            GestName := self.ToHumanReadable(Gest);
            Gestures.Values[GestName] := Gest;
            ActionType := ini.ReadString(Gest, 'ActionType', '');
            ActionTypes.values[Gest] := ActionType;

            if not SkipGui then begin
                li := lvGestures.Items.Add;
                li.caption := GestName;
                li.SubItems.Add(ActionType);
            end;

            if ActionType = ToString(atKeystroke) then  begin
                Action := ini.ReadString(Gest, 'Action', '');
                kd := FrmKeystrokeEdit.FromString(Action);
                Keystrokes.Values[Gest] := FrmKeystrokeEdit.ToString(kd);
                if not SkipGui then li.SubItems.Add(kd.name);

            end else if ActionType = ToString(atMultiKeystroke)  then begin
                display := '';
                Action := ini.ReadString(Gest, 'Action', '');
                s := Action;
                while Action <> '' do begin
                    kd := FrmKeystrokeEdit.FromString( TokenString(Action,'|') );
                    if display <> '' then display := display + ' | ';
                    display := display + kd.name;
                end;
                Keystrokes.values[gest] := s;
                if not SkipGui then li.SubItems.Add(display);
            end else if actiontype = ToString(atProgram)  then begin
                action := ini.readstring(gest, 'Action', '');
                Keystrokes.Values[gest] := action;
                while action <> '' do begin
                    display := TokenString(Action,'|');
                    display := display + ' | ' + TokenString(Action,'|');
                end;
                if not SkipGui then li.SubItems.Add(display);

            end;
        end;
        myfree(ini);
    end;

    if filename = 'Gestures.ini' then begin
        GlobalKeystrokes.Clear;
        GlobalGestures.clear;
        GlobalActionTypes.clear;
        GlobalKeystrokes.AddStrings(Keystrokes);
        GlobalGestures.AddStrings(Gestures);
        GlobalActionTypes.AddStrings(ActionTypes);
    end;
end;



procedure TFrmGestures.lvGesturesClick(Sender: TObject);
begin
    try
        if lvGestures.selected = nil then EXIT;
        FrmDrawGesture.GetImage(Gestures.values[lvGestures.Selected.Caption], self.imgPreview.Canvas,self.imgPreview.Width,self.imgPreview.Height); 
    except

    end;
end;

procedure TFrmGestures.SaveGestures(filename : string = 'Gestures.ini');
var i : integer;
    name, s, gest : string;
    ini : TIniFile;
begin
    name := UnitMisc.GetAppPath + filename;

    ini := TIniFile.Create(name);



    // only save if the list isn't empty
    // or if there's a previous file
    if (lvGestures.Items.count > 0) or (FileExists(name)) then begin
        DeleteFile(name);


        for i := 0 to Gestures.count - 1 do begin
            gest := Gestures.ValueFromIndex[i];
            ini.WriteString(gest, 'ActionType', ActionTypes.Values[gest] );
            case FromString(ActionTypes.Values[gest]) of
                atKeystroke, atMultiKeystroke, atProgram : begin
                    ini.WriteString(gest,'Action', Keystrokes.Values[gest] );
                end;
            end;
        end;
{       for i := 0 to lvGestures.Items.count - 1 do begin
            ini.WriteString(Gestures.values[lvGestures.Items[i].caption], 'ActionType', lvGestures.Items[i].SubItems[0]);
            case FromString(lvGestures.Items[i].SubItems[0]) of
                atKeystroke: begin
                    ini.WriteString(Gestures.values[lvGestures.Items[i].caption], 'Action',  Keystrokes.values[lvGestures.Items[i].SubItems[1]]);
                end;
                atMultiKeystroke: begin
                    ini.WriteString(Gestures.values[lvGestures.Items[i].caption], 'Action', Keystrokes.values[lvGestures.Items[i].SubItems[1]]);
                end;
            end;
        end;
        myfree(ini);  }
    end;
    ini.WriteBool('Settings',cbDisableAll.Name, cbDisableAll.checked);
    myfree(ini);

    if filename = 'Gestures.ini' then begin
        GlobalKeystrokes.Clear;
        GlobalGestures.clear;
        GlobalActionTypes.Clear;
        GlobalKeystrokes.AddStrings(Keystrokes);
        GlobalGestures.AddStrings(Gestures);
        GlobalActionTypes.AddStrings(ActionTypes);
    end;
end;


procedure TFrmGestures.LoadGestureFromWizard(NewItem : boolean = true);
var li : TListItem;
    gestname, gesture : string;
    data, display : string;

    kd : TKeystrokeData;
begin
        if NewItem then begin
            li := lvGestures.Items.Add;
        end else begin
            li := lvGestures.Selected;
        end;

        li.caption := FrmGestureWizard.GestureName;
        if NewItem then begin
            li.SubItems.Add(self.toString( FrmGestureWizard.ActionType ));
        end else begin
            li.SubItems[0] := self.toString( FrmGestureWizard.ActionType );
        end;
        //li.SubItems.Add('none');
        li.Selected := true;
        Gestures.Values[FrmGestureWizard.GestureName] := FrmGestureWizard.Gesture;

        ActionTypes.values[FrmGestureWizard.Gesture] := self.toString(FrmGestureWizard.ActionType);
        Keystrokes.Values[FrmGestureWizard.Gesture] := FrmGestureWizard.ActionData;
        data := FrmGestureWizard.ActionData;
        case FrmGestureWizard.ActionType of
            atKeystroke, atMultiKeystroke : begin
                while data <> '' do begin
                    kd := FrmGestureWizard.FrameKeystrokeEdit1.FromString( TokenString(data,'|') );
                    if display <> '' then display := display + ' | ';
                    display := display + kd.name;
                end;
                if NewItem then begin
                    li.SubItems.Add( display );
                end else begin
                    li.SubItems[1] := display;
                end;
            end;
            atProgram: begin
                if NewItem then begin
                    li.SubItems.Add( TokenString(data,'|') + ' | ' + data)
                end else begin
                    li.SubItems[1] := TokenString(data,'|') + ' | ' + data
                end;

            end;
        end;

end;

procedure TFrmGestures.SpeedButton1Click(Sender: TObject);
begin
    //pmNew.Popup(mouse.cursorpos.x+5, mouse.cursorpos.y+7);
    if FrmGestureWizard.NewGesture then begin
        self.LoadGestureFromWizard;
    end;
end;

procedure TFrmGestures.SpeedButton2Click(Sender: TObject);
begin
    self.Hide;
end;

procedure TFrmGestures.SpeedButton3Click(Sender: TObject);
begin
    FrmConfig.ShowPage(-1);
end;

procedure TFrmGestures.SuggestionConfigureClick(Sender: TObject);
begin
    FrmConfig.ShowPage(3);
end;

procedure TFrmGestures.timCloseTooltipTimer(Sender: TObject);
begin
    timCloseTooltip.Enabled := false;
    tt.CloseTooltip(true);
end;

procedure TFrmGestures.timHelpConfigTimer(Sender: TObject);
begin
    timHelpConfig.Enabled := false;

    FrmConfig.ShowPage(1);
end;

function TFrmGestures.ToHumanReadable(s: string): string;
var i : integer;
begin
    result := '';
    for i := 1 to length(s) do begin
        if result <> '' then result := result + '+';
        result := result + Uppercase( FrmMouseMonitor.ToString(TMouseDirection(StrToInt(s[i]))) );
    end;
end;



procedure TFrmGestures.GestMenuClick(Sender: TObject);
begin
    //Application.ProcessMessages;
    //FrmScratchpad.RemoveSelfFromFocus;
    //Application.ProcessMessages;
    sleep(100);
    ReportGestures( TMenuItem(sender).hint, true );
end;

procedure TFrmGestures.HelpConfigureClick(Sender: TObject);
begin
    timHelpConfig.Enabled := true;
end;

procedure TFrmGestures.NewKeystroke1Click(Sender: TObject);
var s, gest : string;
    li : TListItem;
    kd : TKeystrokeData;
begin
    if FrmGestureEdit.GetGesture(s, gest) then begin
        if Gestures.values[s] = '' then begin

            li := lvGestures.Items.Add;
            li.caption := s;
            li.SubItems.Add(self.toString(atKeystroke));
            li.SubItems.Add('none');
            li.Selected := true;
            Gestures.Values[s] := gest;
            ActionTypes.values[gest] := self.toString(atKeystroke);
            Keystrokes.Values[gest] := FrmKeystrokeEdit.ToString(kd);
        end else begin
            ShowMessage('Gestures already exists');
        end;
    end;
end;

procedure TFrmGestures.Runfileprogram1Click(Sender: TObject);
var s, gest : string;
    li : TListItem;
    kd : TKeystrokeData;
begin
    if FrmGestureEdit.GetGesture(s, gest) then begin
        if Gestures.values[s] = '' then begin

            li := lvGestures.Items.Add;
            li.caption := s;
            li.SubItems.Add(self.toString(atProgram));
            li.SubItems.Add('none');
            li.Selected := true;
            Gestures.Values[s] := gest;
            ActionTypes.values[gest] := self.toString(atProgram);
            Keystrokes.Values[gest] := '';
        end else begin
            ShowMessage('Gestures already exists');
        end;
    end;
end;

procedure TFrmGestures.RunGesture(gesture: string);
var i : integer;
    data,actiontype : string;
    command : string;
    kd : TKeystrokeData;

    function GetEXEPathFromRunValue(value : string) : string;
    var i : integer;
    begin
        i := pos('.exe', lowercase(value));
        result := '';

        if (i > 0) then begin
            result := LeftStr(value, i + 3);

            if leftstr(result,1) ='"' then
                result := RightStr(result, length(result) - 1);
        end;
    end;
    function RunCommandLine(command : string;priority : DWORD = NORMAL_PRIORITY_CLASS; waitfor : boolean = false) : longbool;
    var  StartInfo  : _StartupInfo;
         ProcInfo   : _PROCESS_INFORMATION;
    begin
        // nil for the program name treats the program name
        // as a command line to execute
        FillChar(StartInfo, SizeOf(StartInfo), #0);
        FillChar(ProcInfo, SizeOf(ProcInfo), #0);
        StartInfo.cb := SizeOf(TStartupInfo);
        StartInfo.wShowWindow := SW_HIDE;

        result := CreateProcess(nil, PChar(command),
                    nil, nil, False,
                    CREATE_NEW_PROCESS_GROUP + priority,
                    nil, nil, StartInfo, ProcInfo);

        if (result) then begin
            if waitfor then begin
                WaitForSingleObject(procInfo.hProcess, INFINITE);
            end;
        end else begin
            //ErrorLog.Add(Windows.GetLastError(), command);
        end;
    end;
    function RunCmdAltered(value : string) : boolean;
    var name, params : string;
    begin
        // turn full names into short names and break apart the
        // program and it's parameters

        if (LeftStr(value,1) = '"') then begin
            value := StringReplace(value, '"', '', []);
            value := StringReplace(value, '"', '', []);
        end;

        name := GetEXEPathFromRunValue(value);
        params := StringReplace(value, name, '',[rfIgnoreCase]);

        result := RunCommandLine('"' +  name + '" ' + Trim(params), NORMAL_PRIORITY_CLASS);
    end;

begin
    FrmMouseMonitor.ClosePopups;
    FrmScratchPad.RemoveSelfFromFocus;
    if self.ExecutePause <> 0 then begin
        sleep(self.ExecutePause);
    end;


    //i := Gestures.IndexOf(gesture);
    data := Keystrokes.Values[ gesture ];
    if data = '' then  begin
        data := GlobalKeystrokes.Values[ gesture ];
        actiontype := GlobalActionTypes.Values[ gesture ];

    end else begin
        actiontype := ActionTypes.Values[ gesture ];
    end;
    case fromString(actiontype) of
        atKeystroke, atMultiKeystroke: begin
            while data <> '' do begin
                kd := FrmKeystrokeEdit.FromString( TokenString(data,'|') );
                Paste.SendKestroke(kd);
            end;
        end;
        atProgram: begin
            command := TokenString(data,'|');
            // TODO - handle command line parameters

            command := ExpandVars(command);
            if data  <> '' then begin
                // parameters require extra processing
                command := command + ' '+ TokenString(data,'|');
                RunCommandLine(command);
            end else begin
                command := '[KEYS][RUN]' + command ;
                Paste.SendTextWithKeystrokes(command);
            end;
        end;
    end;
end;

procedure TFrmGestures.MultiKeystrokesTrigger1Click(Sender: TObject);
var s, gest : string;
    li : TListItem;
    kd : TKeystrokeData;
begin
    if FrmGestureEdit.GetGesture(s, gest) then begin
        if Gestures.values[s] = '' then begin

            li := lvGestures.Items.Add;
            li.caption := s;
            li.SubItems.Add(self.toString(atMultiKeystroke));
            li.SubItems.Add('none');
            li.Selected := true;
            Gestures.Values[s] := gest;
            ActionTypes.values[gest] := self.toString(atMultiKeystroke);
            Keystrokes.Values[gest] := FrmKeystrokeEdit.ToString(kd);
        end else begin
            ShowMessage('Gestures already exists');
        end;
    end;
end;


function TFrmGestures.toString(at: TActionType): string;
begin
    case at of
    atKeystroke: result := 'Keystroke';
    atMultiKeystroke: result := 'Multiple Keystrokes';
    atProgram: result := 'Run Program / File';
    end;
end;

function TFrmGestures.fromString(s: string): TActionType;
begin
    if s = 'Keystroke' then begin
        result := atKeystroke;
    end else if s='Multiple Keystrokes' then begin
        result := atMultiKeystroke;
    end else if s = 'Run Program / File' then begin
        result := atProgram;
    end;

end;

end.
